/*
MatrixLib.h - MatrixObj library V1.00
V1.00 2023 Yves Cordier 

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

library created Yves Cordier the 28 march 2023

*/


/* this file is the header of the MatrixObj bibliotheque
it contains the MatrixObj class where nearly all functions are implemented as propieties
and a set of functions (equivalent) which act on standard arrays of real

MatrixObj class
manipulation standard of numerical MatrixObj

multiply addition multiply by number invese transpose ....
and a set of propieties to solve systems by différents methodes

Gauss jordan elimination
LU Crout transformation
classic déterminant methode

*/

/* 
you have to precize the tag used for float either standard float on 4 bytes or double on 8 bytes
be aware that on uno and mega (8 bits processors with few memory), double is interpreted as float.
i.e. on these targets it is the same
*/


#ifndef __Matrixlib_h__
#define __MatrixLib_h__

// as to be defined to set up the size or the real
// note : on uno and mega double does not exist in reality, it is coded at the same size of float
//
#define real float
// #define real double
#define _Epsilon 1E-8

class MatrixObj {
private:
	int sizex = 0;
	int sizey = 0;
	int sizeLin = 0;
	real* contents = nullptr;
public:
	
	// show the MatrixObj on a serial beam
	void showMat(HardwareSerial *ser=&Serial);
	
	//constructor 
	MatrixObj(int l, int c);
	
	// destructor needed to freeing  internal structure created on memory
	~MatrixObj();

	// empty the MatrixObj (set with 0)
	void empty();
	
	// to eliminate small values default _epsilon
	void zeroed(real v=_Epsilon);
	
	// create identity MatrixObj
	boolean identity();
	
	// compute transpose of MatrixObj A=A^T (square part of MatrixObj)
	void transpose();
	
	// multiply MatrixObj by a real A=v*A
	void multiplyVal(real val);
	
	// inverted the MatrixObj  A^-1=1/det(A)*C^T where C is the comatrix of A
	// ci,j=(-1)^(i+j)det(M) where M is the MatrixObj formed by terms of inital MatrixObj exept line i and column j
	// M is a MatrixObj (sise-1 x size-1)	
	boolean invMatrixClassic();
	
	// A=AxM
	// return in self result of multiplication of self by M
	// only the possible parts able to multiply are treated;
	boolean multiply(MatrixObj M);
	
	// A=A+M
	// return in self result of addition of self by M
	// only the commun part of two MatrixObj are treated
	boolean add(MatrixObj M);
	
	// return in self a copy of M (M ans self should be the same sizes)
	boolean copy(MatrixObj M);
	
	// to manipulate MatrixObj component
	// get a pointer to MatrixObj real storage
	real* getContents(){return contents;};
	
	// return sizex of storage
	int getSizeX() {return sizex;};
	
	// return sizey of storage
	int getSizeY() {return sizey;};
	
	// return total size of storage  sizex*sizey 
	int getSizelin() {return sizeLin;};
	
	// how to get a value from MatrixObj return 0 if out of range [0..sizey-1,0..sizex-1]
	real getValMat(int line, int column);

	// to set a value to the MatrixObj return false if out of range[0..sizey-1,0..sizex-1]
	boolean setValMat(int line, int column, real V);

	// to get a value of the system's solution (if exist)  
	// return 0 if does not exist or out of range[0..sizey-1] 
	real getValSol(int line);

	// to set a value of constant vector (n indicate which constant if many from 0 to sizex-sizey-2...)
	// return false if does not exist or out of range[0..sizey-1] and n<sizex-sizey-1 
	boolean setValConst(int n,int l, real V);
	
	// sets the column of solution to 0 if possible
	void clearSolution();
	
	// functions using classis determinant
	// compute determinant
	real deterClassic();
	// solve system with column constant
	real deterSolve(int constColumn=0);

	// functions using gauss jordan methode
	// compute determinant 
	real gaussJordanDeter();
	// transform matrix in upper triangle one (1 on diagonal)
	real gaussJordanPiv(int constColumn=0);
	// solves system in final wit column constant
	void gaussJordanFinalSolve(int constColumn=0);
	// transform and solve
	real gaussJordanSolve(int constColumn=0);

	// LU Crout functions
	// compute determinant
	real luCroutDeter();
	// transform matrix in LU Crout mode
	real luCroutPiv(int constColumn=0);
	// solve system with column constant
	void luCroutFinalSolve(int constColumn=0);
	// transform and solve
	real luCroutSolve(int constColumn=0);

};  // end of class

// *****************************************************************
// all functions out of a class
// Theeses functions are équivalent but they have to be
// called directely

// M1=M1+M2
// only the possible parts able to multiply are treated;
boolean addMat(real *M1,int sizey1,int sizex1,real *M2, int sizey2, int sizex2);

// M1=M1xM2
// only the possible parts able to multiply are treated;
boolean multiplyMat(real *M1,int sizey1,int sizex1,real *M2, int sizey2, int sizex2); 

// Show a MatrixObj on Serial beam
// in entry : The MatrixObj (an array), the number of lines, the number of column
// the serial beam to be used example : showMat(M,4,6,Serial);
void showMatMat(real* M, int sizey,int sizex,HardwareSerial *ser=&Serial);


// show a square MatrixObj on Serial beam
void showSquareMat(real* M, int size,HardwareSerial *ser=&Serial); 

// show a vector on Serial beam
void showVector(real* M, int size,HardwareSerial *ser=&Serial);

// compute classic déterminant
// input M a rectangular MatrixObj sizey lines sizex columns
// only the square part is trated. The MatrixObj became unchanged
// if sizex=-1 (default value), the MatrixObj is taken as square (sizex=sizey)
real deterClassicMat(real* M, int sizey, int sizex=-1);

// inverted the MatrixObj  M^-1=1/det(M)*C^T where C is the comatrix of M
// ci,j=det(CO) where CO is the MatrixObj formed by terms of inital MatrixObj exept line i and column j
// CO is a MatrixObj (sise-1 x size-1)
// input M a rectangular MatrixObj sizey lines sizex columns
// only the square part is inverted. The rest of MatrixObj became unchanged
// if sizex=-1 (default value), the MatrixObj is taken as square (sizex=sizey)
boolean invMatrixClassicMat(real* M, int sizey, int sizex=-1); 


// solve the system MX=B with the kramer méthode
// return determinant of M (if 0 the solution does not exist)
// due to large amount of memory used and
// huge number of opérations, this function is
// only for purist and realy NOT to USE in real problems
// M is a rectangular MatrixObj 
// M contains M,Bi, X M is a rectangular MatrixObj
// constant column indicate the constant column to take constant column from 0 to ...
// last column will store the solution vector
// asumes sizex>sizey+2 (in order to store at least 1 constant column and a vector solution)
real deterSolveMat(real* M, int sizey, int sizex, int constantColumn);

// compute determinant by gauss jordan elimination
// return 0 if problem
real gaussJordanDeterMat(real* M, int sizey, int sizex);

// Gauss Jordan transformation with partial pivoting
// System M X = Bi all store in one MatrixObj
// M is a rectangular MatrixObj
// M contains M,Bi, X M is a rectangular MatrixObj
// constant column indicate the constant column to take constant column from 0 to ...
// asumes sizex>sizey+2 (in order to store at least 1 constant column and a vector solution)
real gaussJordanPivMat(real* M, int sizey, int sizex,int constantColumn);

// solve system  by backward computing asuming M is an upper triangle MatrixObj
// added with constant Bi folowed by a X
void gaussJordanFinalSolveMat(real* M, int sizey, int sizex, int constantColumn);

// solves the systemn MX=Bi M a rectangular MatrixObj wich contain M, Bi, X
// return determinant of square part of M (if 0 the solution does not exist)
real gaussJordanSolveMat(real* M, int sizey, int sizex, int constantColumn); 

// compute determinant by lu Crout pivoting elimination
// return 0 if problem
real luCroutDeterMat(real* M, int sizey, int sizex);

// make a LUosition with no pivoting
// L : lower triangular MatrixObj
// U : upper triangular MatrixObj with 1 on diagonal
// LU=M
// return determinant of M
// pay attention that osition could crash
// due to momentary 0 on diagonal
real luCroutMat(real* M, int sizey, int sizex=-1);

// make a LUosition with  partial pivoting
// LU Crout transformation with partial pivoting
// System M X = Bi all store in one MatrixObj
// M is a rectangular MatrixObj
// M contains M,Bi, X M is a rectangular MatrixObj
// constant column indicate the constant column to take constant column from 0 to ...
// asumes sizex>sizey+2 (in order to store at least 1 constant column and a vector solution)
// L : lower triangular MatrixObj
// U : upper triangular MatrixObj with 1 on diagonal
// LU=M
// B : constant second member of the equation M X = B
// return determinant of M
// pay attention that after transformation deter LU may differ from deter Minit by sign due to 
// line swapping
// return determinant of initial MatrixObj
real luCroutPivMat(real* M, int sizey,int sizex,int constantColumn=-1);

// solves the system M X = Bi
// assuming M is in LU Crout mode
// L is lower triangle
// U is upper triangle with 1 on diagonal
// it gives solution of multiples constant vectors stored with the MatrixObj and
// treated in the same time as MatrixObj itself
// constantColumn varies from 0 to (sizex-sizey-2)
void luFinalSolveMat(real * M, int sizey, int sizex, int constantColumn);

// solves the systemn MX=Bi M a rectangular MatrixObj wich contain M, Bi, X
// return determinant of square part of M (if 0 the solution does not exist)
real luCroutSolveMat(real* M, int sizey, int sizex, int constantColumn);
#endif
